package it.openutils.spring.rmibernate.server.aspects;

import it.openutils.spring.rmibernate.server.aspects.util.EntitySerializer;
import it.openutils.spring.rmibernate.shared.LazyReference;
import it.openutils.spring.rmibernate.shared.managers.HibernateLazyService;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.cglib.proxy.Callback;
import net.sf.cglib.proxy.CallbackFilter;
import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.Factory;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;
import net.sf.cglib.proxy.NoOp;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.collections.SetUtils;
import org.apache.commons.collections.Transformer;
import org.hibernate.EntityMode;
import org.hibernate.SessionFactory;
import org.hibernate.collection.PersistentCollection;
import org.hibernate.collection.PersistentSet;
import org.hibernate.proxy.HibernateProxy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.ReflectionUtils;


/**
 * Proxy object with {@link EntitySerializer} and Intercept writeReplace calls
 * @author mmolaschi
 * @version $Id: SerializationInterceptor.java 485 2007-10-26 11:06:52Z fcarone $
 */
public class SerializationInterceptor implements MethodInterceptor
{

    private static ThreadLocal<List<Object>> processed = new ThreadLocal<List<Object>>();

    private static ThreadLocal<HibernateLazyService> hibernateLazyService = new ThreadLocal<HibernateLazyService>();

    private static Map<Class, Object> proxies = Collections.synchronizedMap(new HashMap<Class, Object>());

    /**
     * log
     */
    private static Logger log = LoggerFactory.getLogger(SerializationInterceptor.class);

    /**
     * non proxied object
     */
    private Object original;

    /**
     * hibernate sessionfactory
     */
    private SessionFactory sessionFactory;

    /**
     * Constructor
     * @param o object to be proxied
     * @param sessionFactory hibernate sessionfactory
     */
    private SerializationInterceptor(Object o, SessionFactory sessionFactory)
    {
        this.original = o;
        this.sessionFactory = sessionFactory;
    }

    /**
     * Cleans the processed beans cache
     */
    public static void clean()
    {
        processed.set(null);
    }

    /**
     * Get proxy object intercepting writeReplace calls
     * @param o object to be proxied
     * @param sessionFactory hibernate sessionfactory
     * @return proxied object
     */
    public static Object getEnhancedObject(Object o, SessionFactory sessionFactory)
    {
        return getEnhancedObject(o, sessionFactory, null);
    }

    /**
     * Get proxy object intercepting writeReplace calls
     * @param o object to be proxied
     * @param sessionFactory hibernate sessionfactory
     * @param lazyRef reference to a lazy field
     * @return proxied object
     */
    public static Object getEnhancedObject(Object o, SessionFactory sessionFactory, LazyReference lazyRef)
    {
        if (o == null)
        {
            return null;
        }

        if (!(o instanceof HibernateProxy)
            && !(o instanceof PersistentSet)
            && processed.get() != null
            && processed.get().contains(o))
        {
            return o;
        }

        if (o.getClass().getName().startsWith("java.") && !((o instanceof Collection) || (o instanceof Map)))
        {
            return o;
        }

        // check if there is an empty constructor
        try
        {
            o.getClass().getConstructor(new Class[]{});
        }
        catch (NoSuchMethodException ex)
        {
            return o;
        }

        // check if object can be subclassed
        if (Modifier.isFinal(o.getClass().getModifiers()))
        {
            return o;
        }

        try
        {
            // get class
            Class clazz = o.getClass();

            // if it is an hibernateproxy get superclass
            if (o instanceof HibernateProxy)
            {
                clazz = o.getClass().getSuperclass();
                if (clazz == null)
                {
                    clazz = o.getClass().getInterfaces()[0];
                }
            }

            Callback callback = null;

            // if this is a lazy field user lazyreferenceaspect
            if (lazyRef != null)
            {
                LazyReferenceAspect lra = new LazyReferenceAspect();
                lra.setLazyReference(lazyRef);
                callback = lra;
            }
            else
            {
                // user default interceptor
                callback = new SerializationInterceptor(o, sessionFactory);
            }

            synchronized (proxies)
            {
                if (proxies.containsKey(clazz))
                {
                    Factory proxy = (Factory) proxies.get(clazz);
                    return proxy.newInstance(new Callback[]{callback, NoOp.INSTANCE });
                }
            }
            // create proxy to listen on writeReplace method calls
            Object proxy = Enhancer.create(clazz, new Class[]{EntitySerializer.class }, new CallbackFilter()
            {

                public int accept(Method method)
                {
                    if (method.getName().equals("writeReplace"))
                    {
                        return 0;
                    }
                    else
                    {
                        return 1;
                    }
                }

            }, new Callback[]{callback, NoOp.INSTANCE });

            // store proxy
            proxies.put(clazz, ((Factory) proxy).newInstance(new Callback[]{
                EmptyMethodInterceptor.INSTANCE,
                NoOp.INSTANCE }));
            return proxy;
        }
        catch (Throwable t)
        {
            log.warn(t.getMessage(), t);
            return o;
        }
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Object intercept(Object obj, Method method, Object[] args, MethodProxy proxy) throws Throwable
    {
        try
        {
            if (original == null)
            {
                return null;
            }

            if (original.getClass().getName().startsWith("java.")
                && !(original instanceof Collection)
                || (original instanceof Map))
            {
                return original;
            }

            if (processed.get() == null)
            {
                processed.set(new ArrayList<Object>());
            }

            // add to processed objects
            processed.get().add(original);

            if (original.getClass().isPrimitive())
            {
                log.debug("Non enhancing primitive type: {}", original.getClass().getName());
                return original;
            }
            else if (original.getClass().isArray())
            {
                // replace array entries with proxied ones
                Object[] array = (Object[]) original;
                for (int i = 0; i < array.length; i++)
                {
                    array[i] = SerializationInterceptor.getEnhancedObject(array[i], sessionFactory);
                }
                return array;
            }
            else if (original instanceof Set)
            {
                // replace set entries with proxied ones
                Set set = (Set) original;
                SetUtils.transformedSet(set, transformer(sessionFactory));
                return set;
            }
            else if (original instanceof Collection)
            {
                // replace collection entries with proxied ones
                Collection collection = (Collection) original;
                CollectionUtils.transform(collection, transformer(sessionFactory));
                return collection;
            }
            else if (original instanceof Map)
            {
                // replace map entries with proxied ones
                Map map = (Map) original;
                MapUtils.transformedMap(map, transformer(sessionFactory), transformer(sessionFactory));
                return map;
            }
            else
            {
                // cycle on bean fields
                ReflectionUtils.doWithFields(original.getClass(), new ReflectionUtils.FieldCallback()
                {

                    /**
                     * {@inheritDoc}
                     */
                    public void doWith(Field field) throws IllegalArgumentException, IllegalAccessException
                    {
                        // set field accessible
                        field.setAccessible(true);

                        // get original value
                        Object oldValue = field.get(original);

                        // get new value
                        Object newValue = null;

                        if (oldValue instanceof HibernateProxy || oldValue instanceof PersistentCollection)
                        {
                            // if this field is a hibernate proxy or a persistent collection store reference
                            LazyReference lazyRef = new LazyReference();
                            lazyRef.setFieldClassName(field.getType().getName());
                            lazyRef.setClassName(original.getClass().getName());
                            lazyRef.setFieldName(field.getName());
                            // load id
                            lazyRef.setId(sessionFactory.getClassMetadata(original.getClass()).getIdentifier(
                                original,
                                EntityMode.POJO));

                            // get proxy for lazy
                            newValue = SerializationInterceptor.getEnhancedObject(oldValue, sessionFactory, lazyRef);
                        }
                        else
                        {
                            // get default proxy
                            newValue = SerializationInterceptor.getEnhancedObject(oldValue, sessionFactory);
                        }

                        // if there is a variation store it
                        if (newValue != oldValue)
                        {
                            field.set(original, newValue);
                        }
                    }

                });
            }

            return original;

        }
        catch (Throwable t)
        {
            log.error(t.getMessage(), t);
            return original;
        }
    }

    /**
     * Get transformer for collection, map or set injection of proxied objects
     * @param factory hibernate session factory
     * @return transformer
     */
    private Transformer transformer(SessionFactory factory)
    {
        return new Transformer()
        {

            public Object transform(Object input)
            {
                return SerializationInterceptor.getEnhancedObject(input, sessionFactory);
            }

        };
    }

    /**
     * Sets the hibernateLazyService.
     * @param hibernateLazyService the hibernateLazyService to set
     */
    public static void setHibernateLazyService(HibernateLazyService hibernateLazyService)
    {
        SerializationInterceptor.hibernateLazyService.set(hibernateLazyService);
    }

    /**
     * Returns the hibernateLazyService.
     * @return the hibernateLazyService
     */
    public static HibernateLazyService getHibernateLazyService()
    {
        return hibernateLazyService.get();
    }
}
